package MultiCastZoo;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.TreeMap;

/**
 * Zoo -- holds one zoo in a distributed zoo system.
 * Written for a computer security exercise in
 * Computer Network Security (ECE 6612 Spring 2008)
 * at Georgia Tech
 * Integrated into MultiCastZoo by David Bauer and Chris Lee.
 * @author David Bauer
 */
public class Zoo {
	static Random rand = new Random();
	private String identity = "Zoo" + (Math.abs(rand.nextInt() % 100000));
	private Map<String, Species> animals;
	private Map<String, Map<String, Integer>> neighbors;
	private List<String> outMsgQueue = new LinkedList<String>();

	public Zoo(String configFile) throws IOException {
		readFromConfigFile(configFile);
		neighbors = new TreeMap<String, Map<String, Integer>>();
	}

	public static final class Species {
		String name;
		int population;
		int maxPopulation;
		double breedingRate;		/* Probability of each two animals producing a new animal each turn. */
		double lifespan;			/* Probability of each animal dying each turn. */
		double travel;				/* Biases probability of traveling versus staying still */
		List<NameProb> eating;
		List<NameProb> chases;		/* List of species with bias for chasing (positive values) or avoid (negative values) */
		int feed;
	}

	public static final class NameProb {
		String name;
		double prob;

		public NameProb(String n, double p) {
			name = n;
			prob = p;
		}
	}

	public String getIdentity() {
		return identity;
	}

	public void addNeighbor(String name) {
		neighbors.put(name, new TreeMap<String, Integer>());
	}

	public String getAll() {
		String str;
		str = (identity + " " + animals.size() + " species and " + neighbors.size() + " neighbors\n");
		for (Species group : animals.values()) {
			str += group.name + " : " + group.population + " (" + group.maxPopulation + "), " + group.breedingRate +
					", " + group.lifespan + ", " + group.travel + "\n";
		}
		for (String key : neighbors.keySet()) {
			Map<String, Integer> neighbor = neighbors.get(key);
			str += key + " : ";
			for (Map.Entry<String, Integer> entry : neighbor.entrySet()) {
				str += "[" + entry.getKey() + " : " + entry.getValue() + "]";
			}
			str += "\n";
		}
		return str;
	}

	private void readFromConfigFile(String configFile) throws IOException {
		BufferedReader in = new BufferedReader(new FileReader(configFile));
		animals = new TreeMap<String, Species>();
		while (true) {
			int i;
			String line = in.readLine();
			if (line == null) break;	/* End of file */
			line = line.trim();
			if (line.startsWith("#")) continue;		/* Skip comment lines */
			/* Format should be:
			 * <Species name> <initial population> <max population> <breeding rate> <inverse lifespan> [<eating habits>]
			 * <eating habits> is done as <species name> <prob to eat per animal> <species name> <prob to eat per animal ...
			 */
			/* TODO:  have a group for "eats" and a group for "chases", and maybe a group for "avoids" */
			Species group = new Species();
			String [] parts = line.split(" ");
			for (i = 0; i < parts.length; i++) System.out.print("\"" + parts[i] + "\" ");
			System.out.println("\nTotal of " + parts.length + " parts");
			if (parts.length < 2) continue;		// Empty line
			if (parts.length < 5) throw new IOException("Configuration file format was invalid");
			group.name = parts[0];
			group.population = Integer.parseInt(parts[1]);
			group.maxPopulation = Integer.parseInt(parts[2]);
			group.breedingRate = Double.parseDouble(parts[3]);
			group.lifespan = Double.parseDouble(parts[4]);
			group.eating = new LinkedList<NameProb>();
			group.chases = new LinkedList<NameProb>();
			int pos = 5;
			if (pos < parts.length && parts[pos].equalsIgnoreCase("eats")) {
				for (i = pos + 1; i < parts.length - 1; i+=2) {
					if (parts[i].equalsIgnoreCase("chases")) break;
					NameProb eats = new NameProb(parts[i], Double.parseDouble(parts[i+1]));
					group.eating.add(eats);
				}
				pos = i;
			}
			if (pos < parts.length && parts[pos].equalsIgnoreCase("chases")) {
				for (i = pos + 1; i < parts.length - 1; i+= 2) {
					if (parts[i].equalsIgnoreCase("travels")) break;
					NameProb chases = new NameProb(parts[i], Double.parseDouble(parts[i+1]));
					group.chases.add(chases);
				}
				pos = i;
			}
			if (pos < parts.length && parts[pos].equalsIgnoreCase("travels")) {
				group.travel = Double.parseDouble(parts[pos + 1]);
			}
			animals.put(group.name, group);
		}
		in.close();
	}

	private int getPopulation(String name) {
		return animals.get(name).population;
	}

	private double getAttraction(String species, String neighbor) {
		double score = 0;
		if (neighbor.equals(identity)) {
			/* TODO  Local zoo! */
			/* Ignoring it seems to work okay. */
		}
		Species group = animals.get(species);
		Map<String, Integer> zoo = neighbors.get(neighbor);
		if (group.chases.size() == 0) score = 1;
		else {
			for (NameProb chase : group.chases) {
				if (zoo.get(chase.name) == null) return 0.0001;	/* simple epsilon */
				int pop = zoo.get(chase.name);
				score += pop * chase.prob;
			}
		}
		if (!neighbor.equals(identity)) score *= group.travel / neighbors.size();
//		System.out.println("Debug: " + species + " ** Score = " + score);
		return score;
	}

	/**
	 * Find the relative probability of species wanting to travel to 
	 * @param species
	 * @return
	 */
	private List<NameProb> findDestProbs(String species) {
		List<NameProb> neighList = new LinkedList<NameProb>();
		double min = Integer.MAX_VALUE;
		double max = Integer.MIN_VALUE;
		double sum = 0;
		for (String neighbor : neighbors.keySet()) {
			NameProb entry = new NameProb(neighbor, getAttraction(species, neighbor));
			if (entry.prob < min) min = entry.prob;
			if (entry.prob > max) max = entry.prob;
			sum += entry.prob;
			neighList.add(entry);
		}
		if (min > 0) min = 0;		// Don't scale min to zero; only care about negatives
//		System.out.println("Debug: " + species + " sum = " + sum + ", list size = " + neighList.size());
		sum -= min * neighList.size();			/* Compensate for shifting toward zero */
		double scale = sum; // / neighList.size();	/* Find scaling factor */
		for (NameProb entry : neighList) {		/* Scale all of the entries such that their sum is 1. */
			entry.prob = (entry.prob - min) / scale;
		}
		return neighList;
	}

	public String [] performRound() {
		try {
			breed();
			eat();
			age();
			migrate();
			return updateOthers().toArray(new String[0]);
		} catch (Exception e) {
			System.out.println(e);
			return new String[0];
		}
	}
	// In each round: breed, eat, die of old age, and send away animals, broadcast status (may be combined with send away animals)
	public void breed() {
		for (Species group : animals.values()) {
			 int delta = (int) ((group.feed / 2) * group.breedingRate * rand.nextDouble() * 4);
			 group.population += delta;
			 if (group.population > group.maxPopulation) group.population = group.maxPopulation;
//			 System.out.println("Debug: " + delta + " new " + group.name + " were born, now " + group.population);
		}
	}

	public void eat() {
		for (Species group : animals.values()) {
			group.feed = 0;
			if (group.eating.size() == 0) group.feed = group.population;
			for (int i = 0; i < group.population; i++) {	// Doing an individual by individual loop should reduce finishing off a population
				double target = rand.nextDouble();
				for (NameProb eating : group.eating) {
					target -= eating.prob * getPopulation(eating.name);
					if (target < 0) {
						animals.get(eating.name).population--;	// We ate an animal!
						group.feed++;
//						System.out.println("Debug: A " + group.name + " ate a " + eating.name);
						break;		// Only eat one thing per round
					}
				}
			}
		}
	}

	public void age() {
		for (Species group : animals.values()) {
			int delta = (int) (group.population * group.lifespan * rand.nextDouble());	/* Old age */
			delta += (int) ((group.population - group.feed) * group.lifespan);				/* Starvation */
			if (group.population - delta > group.maxPopulation)
				delta += group.population - delta - group.maxPopulation;		/* Over population */
			if (group.population - delta < 0)			/* No over-kills */
				delta = group.population;
			group.population -= delta;
			if (group.population < 0) {
				System.out.println("Error:  Tried to decrement population below zero!");
				System.out.println(group.name + ", delta = " + delta + ", new pop = " + group.population);
				System.out.println(getAll());
			}
//			System.out.println("Debug: " + delta + " old " + group.name + " died, now " + group.population);
		}
	}

	public void migrate() {
		for (Species group : animals.values()) {
			List<NameProb> destinations = findDestProbs(group.name);
//			for (NameProb np : destinations) System.out.print("(" + np.name + ", " + np.prob + ") ");
//			System.out.println("");
			for (int i = 0; i < group.population; i++) {	// I would prefer not to do individual by individual....
				double target = rand.nextDouble();
				for (NameProb dest : destinations) {
					target -= dest.prob;
					if (target < 0) {
						if (!dest.name.equals(identity)) performMigrate(group, dest);
						break;
					}
				}
			}
		}
	}

	private void performMigrate(Species group, NameProb dest) {
		group.population--;
		String msg = "MIGRATE 1 " + group.name + " to " + dest.name;
		outMsgQueue.add(msg);
	}

	private void performPopUpdate() {
		for (Species group : animals.values()) {
			String msg = "POPULATION " + group.name + " " + group.population + " in " + identity;
			outMsgQueue.add(msg);
		}
	}

	public List<String> updateOthers() {
		performPopUpdate();						/* Population update follow migration; accounts for births and deaths. */
		List<String> ret = outMsgQueue;
		outMsgQueue = new LinkedList<String>();		/* Reset list for next round. */
		return ret;
	}

	public void checkSelfNeighbor() {
		Map<String, Integer> self = neighbors.get(identity);
		for (Species group : animals.values()) {
			if (self.get(group.name) != group.population) {
				System.out.println("ERROR: Population does not match entry in neighbor's table");
				System.out.println(getAll());
				throw new RuntimeException("Error: Population does not match entry in neighbor's table!");
			}
		}
	}

	public void updateSelfNeighbor() {
		Map<String, Integer> self = neighbors.get(identity);
		for (Species group : animals.values()) {
			self.put(group.name, group.population);
		}
	}

	public int countSpecies(String name) {
		int sum = 0;
		for (Map<String, Integer> neighbor : neighbors.values()) {
			if (neighbor.get(name) != null)
				sum += neighbor.get(name);
		}
		return sum;
	}

	public boolean checkFailure() {
		for (Species group : animals.values()) {
			if (countSpecies(group.name) < 1) return true;
		}
		return false;
	}

	public void receiveMessage(String message) {
		try {
			String [] messages =  message.split("\n");
			List<String> list = new LinkedList<String>();
			for (String msg : messages) list.add(msg);
			updateSelf(list);
		} catch (Exception e) {
			System.out.println(e);
		}
	}

	public void updateSelf(List<String> messages) {
		for (String msg : messages) {
			if (msg.startsWith("MIGRATE")) {
				String [] pieces = msg.split(" ");
				if (pieces.length < 5) {
					System.out.println("Received bad message: \"" + msg + "\"");
					continue;
				}
				int num = Integer.parseInt(pieces[1]);
				if (num < 0) {
					System.out.println("Recieved bad message: \"" + msg + "\"");
					continue;
				}
				String name = pieces[2];	String id = pieces[4];
				if (id.equals(identity)) {
					Species group = animals.get(name);
					group.population += num;
					if (group.population > group.maxPopulation) group.population = group.maxPopulation;
				}
				if (neighbors.get(id) != null && neighbors.get(id).get(name) != null) {
					int pop = neighbors.get(id).get(name);
					neighbors.get(id).put(name, pop + num);
				}
			}
			if (msg.startsWith("POPULATION")) {
				String [] pieces = msg.split(" ");
				if (pieces.length < 5) {
					System.out.println("Recieved bad message: \"" + msg +"\"");
					continue;
				}
				int num = Integer.parseInt(pieces[2]);
				if (num < 0) {
					System.out.println("Recieved bad message: \"" + msg + "\"");
					continue;
				}
				String name = pieces[1];
				if (animals.get(name) == null) {
					System.out.println("Unknown species: \"" + name + "\"");
					continue;
				}
				Map<String, Integer> zoo = neighbors.get(pieces[4]);
				if (zoo == null) {
					addNeighbor(pieces[4]);
					System.out.println("Added neighbor: \"" + pieces[4] + "\"");
					continue;
				}
				zoo.put(name, num);
			}
		}
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) throws IOException {
		Zoo [] zoos = new Zoo[5];

		for (int i = 0; i < zoos.length; i++) {
			zoos[i] = new Zoo("zootest1.cfg");
		}

		for (Zoo z : zoos) {
			for (Zoo oz : zoos) {
				z.addNeighbor(oz.getIdentity());
			}
		}

		for (int i = 0; i < 1000; i++) {
			for (Zoo z : zoos) {
				// System.out.println("Starting round of zoo: " + z.getIdentity());
				z.breed();
				z.eat();
				z.age();
				// z.updateSelfNeighbor();
				z.migrate();
				// z.updateSelfNeighbor();
				List<String> msgs = z.updateOthers();
				// for (String msg : msgs) System.out.println(msg);
				for (Zoo oz : zoos) {
					// if (z == oz) continue;	// Try both with and without this
					oz.updateSelf(msgs);
				}
				z.checkSelfNeighbor();
				if (z.checkFailure()) {
					System.out.println("Failed in round " + i);
					System.out.println(z.getAll());
					return;
				}
			}
			if (i % 10 == 0) {
				System.out.println(zoos[0].getAll());
//				System.out.println(zoos[1].getAll());
//				System.out.println(zoos[2].getAll());
			}
		}

		/*
		Zoo z = new Zoo("zootest1.cfg");

		for (int i = 0; i < 70; i++) {
			z.breed();
			z.eat();
			z.age();
		}
		*/
		return;
	}
}
